@@ -123,11 +123,15 @@ char *MODE_SPEEDTEST;
 char *WAIT_MESSAGE;
 char *ERROR_TOO_HIGH_MSG;
 char *SKIPBACK_VIA_F_MSG;
-char *WANNA_REPEAT_MSG;
+char *REPEAT_NEXT_EXIT_MSG;
+char *REPEAT_EXIT_MSG;
+char *CONFIRM_EXIT_LESSON_MSG;
+char *NO_SKIP_MSG;
 char *SPEED_RAW;
 char *SPEED_ADJ;
 char *SPEED_PCT_ERROR;
 char *YN;
+char *RNE;
 
 #ifndef PACKAGE_DATA_DIR
 #define PACKAGE_DATA_DIR "."
@@ -164,7 +168,8 @@ static char	*cl_start_label = NULL;		/* initial lesson start point*/
 static bool	cl_colour = FALSE;		/* set if -c given */
 static int	cl_fgcolour = 0;		/* fg colour */
 static int	cl_bgcolour = 0;		/* bg colour */
-static bool	cl_wpmode = FALSE;		/* do wp-like stuff */
+static bool	cl_wp_emu = FALSE;		/* do wp-like stuff */
+static bool     cl_no_skip = FALSE;             /* forbid the user to */
 
 /* a few global variables */
 static char	*argv0 = NULL;
@@ -192,14 +197,14 @@ static struct label_entry	*global_label_list[NLHASH];
 /* a global area for associating function keys with labels */
 #define NFKEYS			12		/* num of function keys */
 static	char	*fkey_bindings[ NFKEYS ] =
-{ NULL, NULL, NULL, NULL, NULL, NULL,
-  NULL, NULL, NULL, NULL, NULL, NULL };
+  { NULL, NULL, NULL, NULL, NULL, NULL,
+    NULL, NULL, NULL, NULL, NULL, NULL };
 /* table of pseudo-function keys, to allow ^Q to double as Fkey1, etc */
 #define	CTRL_OFFSET		0100		/* ctrl keys are 'X' - 0100 */
 static	char	pfkeys[ NFKEYS ] =
-{ 'Q'-CTRL_OFFSET, 'W'-CTRL_OFFSET, 'E'-CTRL_OFFSET, 'R'-CTRL_OFFSET,
-  'T'-CTRL_OFFSET, 'Y'-CTRL_OFFSET, 'U'-CTRL_OFFSET, 'I'-CTRL_OFFSET,
-  'O'-CTRL_OFFSET, 'P'-CTRL_OFFSET, 'A'-CTRL_OFFSET, 'S'-CTRL_OFFSET };
+  { 'Q'-CTRL_OFFSET, 'W'-CTRL_OFFSET, 'E'-CTRL_OFFSET, 'R'-CTRL_OFFSET,
+    'T'-CTRL_OFFSET, 'Y'-CTRL_OFFSET, 'U'-CTRL_OFFSET, 'I'-CTRL_OFFSET,
+    'O'-CTRL_OFFSET, 'P'-CTRL_OFFSET, 'A'-CTRL_OFFSET, 'S'-CTRL_OFFSET };
 
 
 #define MAX(A,B) ((A)<(B)?(B):(A))
@@ -213,7 +218,7 @@ static void get_script_line( FILE *script, char *line );
 static int hash_label( char *label );
 static void index_labels( FILE *script );
 static void seek_label( FILE *script, char *label, char *ref_line );
-static bool wait_user( char *message, char *mode, FILE *script, char *line );
+static void wait_user( char *message, char *mode );
 static void display_speed( int total_chars, long elapsed_time, int errcount );
 static void do_keybind( FILE *script, char *line );
 static void do_tutorial( FILE *script, char *line );
@@ -224,8 +229,9 @@ static void do_speedtest( FILE *script, char *line );
 static void do_clear( FILE *script, char *line );
 static void do_goto( FILE *script, char *line, bool flag );
 static void do_exit( FILE *script );
-static bool do_query_noscriptfile( FILE *script, const char *prompt );
-static bool do_query( FILE *script, char *line, bool get_next_script_line );
+static char do_query_repeat( FILE *script, bool allow_next );
+static bool do_query_simple( char *text );
+static bool do_query( FILE *script, char *line );
 static void do_error_max_set( FILE *script, char *line );
 static void do_on_failure_label_set( FILE *script, char *line );
 static void parse_file( FILE *script, char *label );
@@ -332,7 +338,7 @@ fatal_error( char *message, char *line ) {
 /*
   get the next non-comment, non-blank line from the script file
   and check its basic format
- */
+*/
 static void 
 get_script_line( FILE *script, char *line ) {
 
@@ -369,7 +375,7 @@ get_script_line( FILE *script, char *line ) {
 
 /*
   label hashing function - returns an index for the lists
- */
+*/
 static int 
 hash_label( char *label ) {
   char	*p;				/* pointer through string */
@@ -384,8 +390,8 @@ hash_label( char *label ) {
 
 
 /*
-   go through the file and index all the labels we can find
- */
+  go through the file and index all the labels we can find
+*/
 static void index_labels( FILE *script ) {
   char line[MAX_SCR_LINE];
   char *line_iterator;
@@ -470,7 +476,7 @@ static void index_labels( FILE *script ) {
 /*
   search out a label from the file, and set the file pointer to
   that location
- */
+*/
 static void 
 seek_label( FILE *script, char *label, char *ref_line ) {
   struct label_entry	*check_label;	/* pointer through list */
@@ -505,15 +511,12 @@ seek_label( FILE *script, char *label, char *ref_line ) {
 
 
 /*
-  wait for a nod from the user before continuing -- returns true if
-  we just got the expected return, false if exit was by a function
-  key
- */
-static bool 
-wait_user( char *message, char *mode, FILE *script, char *line ) {
+  wait for a nod from the user before continuing
+*/
+static void
+wait_user( char *message, char *mode )
+{
   int	resp;			/* response character */
-  int	fkey;			/* function key iterator */
-  bool	ret_code;		/* return code */
 
   /* move to the message line print a prompt */
   move( MESSAGE_LINE, 0 ); clrtoeol();
@@ -522,62 +525,18 @@ wait_user( char *message, char *mode, FILE *script, char *line ) {
   move( MESSAGE_LINE, 0 );
   ADDSTR_REV( message );
 
-  while ( TRUE ) 
-    {
-      resp = getch_fl( ASCII_NULL );
-      
-      /* translate pseudo Fkeys into real ones if applicable 
-	 The pseudo keys are defined in array pfkeys and are also:
-	 F1 - 1, F2 - 2, F3 - 3,.... F10 - 0, F11 - A, F12 - S */
-      for ( fkey = 1; fkey <= NFKEYS; fkey++ ) 
-	{
-	  if ( resp == pfkeys[ fkey - 1 ] || (fkey<11 && resp == (fkey+'0'))
-	       || (fkey==10 && (resp =='0'))
-	       || (fkey==11 && (resp =='a' || resp=='A'))
-	       || (fkey==12 && (resp =='s' || resp=='S'))) 
-	    {
-	      resp = KEY_F( fkey );
-	      break;
-	    }
-	}
-      
-      /* search the key bindings for a matching key */
-      for ( fkey = 1; fkey <= NFKEYS; fkey++ ) 
-	{
-	  if ( resp == KEY_F( fkey )
-	       && fkey_bindings[ fkey - 1 ] != NULL ) 
-	    {
-	      seek_label( script, fkey_bindings[ fkey - 1 ],
-			  NULL );
-	      get_script_line( script, line );
-	      break;
-	    }
-	}
-      if ( fkey <= NFKEYS ) 
-	{
-	  ret_code = FALSE;
-	  break;
-	}
-      
-      /* not an FKEY - just check for return */
-      if ( (char)resp == ASCII_ENTER ) 
-	{
-	  ret_code = TRUE;
-	  break;
-	}
-    }
-  
+  do {
+    resp = getch_fl( ASCII_NULL );
+  } while (resp != ASCII_ENTER);
+
   /* clear the message line */
   move( MESSAGE_LINE, 0 ); clrtoeol();
-  
-  /* tell the caller if we got a nod or a function key */
-  return ( ret_code );
 }
 
 
 /*
   display speed and accuracy from a drill or speed test
- */
+*/
 static void display_speed( int total_chars, long elapsed_time, int errcount ) {
   double	test_time;			/* time in minutes */
   double	words;				/* effective words typed */
@@ -616,7 +575,7 @@ static void display_speed( int total_chars, long elapsed_time, int errcount ) {
 
 /*
   bind a function key to a label
- */
+*/
 static void 
 do_keybind( FILE *script, char *line ) {
   int	fkey;				/* function key number */
@@ -652,7 +611,7 @@ do_keybind( FILE *script, char *line ) {
 
 /*
   print the given text onto the screen
- */
+*/
 static void 
 do_tutorial( FILE *script, char *line ) {
   int	linenum;			/* line counter */
@@ -676,14 +635,14 @@ do_tutorial( FILE *script, char *line ) {
   /* wait for a return, unless the next command is a query,
      when we can skip it to save the user keystrokes */
   if ( SCR_COMMAND( line ) != C_QUERY )
-    wait_user( WAIT_MESSAGE, MODE_TUTORIAL, script, line );
+    wait_user( WAIT_MESSAGE, MODE_TUTORIAL );
   global_prior_command = C_TUTORIAL;
 }
 
 
 /*
   print up a line, at most two, usually followed by a drill or a speed test
- */
+*/
 static void 
 do_instruction( FILE *script, char *line ) {
 
@@ -709,7 +668,7 @@ do_instruction( FILE *script, char *line ) {
 
 /*
   buffer up the complete data from a command; used for D and P
- */
+*/
 static char *buffer_command( FILE *script, char *line ) {
   int	total_chars = 0;		/* character counter */
   char	*data = NULL;			/* data string */
@@ -748,7 +707,7 @@ static char *buffer_command( FILE *script, char *line ) {
 
 /*
   execute a typing drill
- */
+*/
 static void 
 do_drill( FILE *script, char *line ) {
 
@@ -826,7 +785,6 @@ do_drill( FILE *script, char *line ) {
 			     ASCII_TAB : ASCII_SPACE );
 	      c = (char)rc;
 	    }
-	  if ( c == ASCII_ESC ) break;
 	
 	  /* start timer on first char entered */
 	  if ( chars_typed == 0 )
@@ -840,10 +798,15 @@ do_drill( FILE *script, char *line ) {
 	      p--;		/* defeat p++ coming up */
 	      continue;
 	    }
+
+	  /* ESC is "give up"; ESC at beginning of exercise is "skip lesson"
+	     (this is handled outside the for loop) */
+	  if ( c == ASCII_ESC )
+	    break;
 	
 	  /* check that the character was correct */
 	  if ( c == *p
-	       || ( cl_wpmode && c == ASCII_SPACE
+	       || ( cl_wp_emu && c == ASCII_SPACE
 		    && *p == ASCII_NL ))
 	    ADDCH( c );
 	  else 
@@ -866,7 +829,7 @@ do_drill( FILE *script, char *line ) {
 	    }
 
 	  /* perform any other word processor like adjustments */
-	  if ( cl_wpmode ) 
+	  if ( cl_wp_emu ) 
 	    {
 	      if ( c == ASCII_SPACE ) 
 		{
@@ -899,54 +862,62 @@ do_drill( FILE *script, char *line ) {
 		}
 	    }
 	}
-      /* look for escape key to "give up" */
-      if ( c == ASCII_ESC ) continue;
-    
-      /* display timings */
-      end_time = (long)time( NULL );
-      if ( ! cl_notimer ) 
+      
+      /* ESC not at the beginning of the lesson: "give up" */
+      if ( c == ASCII_ESC && chars_typed != 1)
+	continue; /* repeat */
+
+      /* skip timings and don't check error-pct if exit was through ESC */
+      if ( c != ASCII_ESC )
 	{
-	  display_speed( chars_typed, end_time - start_time,
-			 errors );
+	  /* display timings */
+	  end_time = (long)time( NULL );
+	  if ( ! cl_notimer ) 
+	    {
+	      display_speed( chars_typed, end_time - start_time,
+			     errors );
+	    }
+
+	  /* check whether the error-percentage is too high (unless in d:) */
+	  if (drill_type != C_DRILL_PRACTICE_ONLY &&
+	      (float)errors/(float)chars_typed > global_error_max/100.0) 
+	    {
+	      sprintf( message, ERROR_TOO_HIGH_MSG, global_error_max );
+	      wait_user( message, MODE_DRILL );
+
+	      /* check for F-command */
+	      if (global_on_failure_label != NULL)
+		{
+		  /* move to the label position in the file */
+		  if (fseek(script,global_on_failure_label->offset,SEEK_SET )==-1)
+		    fatal_error( _("internal error: fseek"), NULL );
+		  global_line_counter = global_on_failure_label->line_count;
+		  /* tell the user about the misery :) */
+		  sprintf(message,SKIPBACK_VIA_F_MSG,
+			  global_on_failure_label->label);
+		  /* reset value unless persistent */
+		  if (!global_on_failure_label_persistent)
+		    global_on_failure_label = NULL;
+		  wait_user( message, MODE_DRILL );
+		  seek_done = TRUE;
+		  break;
+		}
+
+	      continue;
+	    }
 	}
 
-      /* check whether the error-percentage is too high (unless in d:) */
-      if (drill_type != C_DRILL_PRACTICE_ONLY &&
-	  (float)errors/(float)chars_typed > global_error_max/100.0) 
-	{
-	  sprintf( message, ERROR_TOO_HIGH_MSG, global_error_max );
-	  if ( ! wait_user( message, MODE_DRILL, script, line )) {
-	    seek_done = TRUE;
-	    break; /* function key */
-	  }
-
-	  /* check for F-command */
-	  if (global_on_failure_label != NULL) {
-	    /* move to the label position in the file */
-	    if (fseek(script,global_on_failure_label->offset,SEEK_SET ) == -1)
-	      fatal_error( _("internal error: fseek"), NULL );
-	    global_line_counter = global_on_failure_label->line_count;
-	    /* tell the user about the misery :) */
-	    sprintf(message,SKIPBACK_VIA_F_MSG,global_on_failure_label->label);
-	    /* reset value unless persistent */
-	    if (!global_on_failure_label_persistent)
-	      global_on_failure_label = NULL;
-	    /* no need to check for fkey because seek_done is set anyway */
-	    wait_user( message, MODE_DRILL, script, line );
-	    seek_done = TRUE;
-	    break;
-	  }
-
-	  continue;
-	}
-
-      /* ask the user whether he/she wants to repeat */
-      if ( ! do_query_noscriptfile ( script, WANNA_REPEAT_MSG ) ) {
+      /* ask the user whether he/she wants to repeat or exit */
+      if ( c == ASCII_ESC && cl_no_skip ) /* honor --no-skip */
+	c = do_query_repeat (script, FALSE);
+      else
+	c = do_query_repeat (script, TRUE);
+      if (c == 'E') {
 	seek_done = TRUE;
-	break; /* function key */
+	break;
       }
-      if (global_resp_flag)
-	break; /* user has hit Y */
+      if (c == 'N')
+	break;
 
     }
   
@@ -957,9 +928,9 @@ do_drill( FILE *script, char *line ) {
   if (!global_error_max_persistent)
     global_error_max = cl_default_error_max;
 
-  /* buffer_command takes care of advancing `script' (and setting `line'),
-     so we only do it if an fkey was pressed (as a side-effect of a query/wait)
-  */
+  /* buffer_command takes care of advancing `script' (and setting
+     `line'), so we only do if seek_label had been called (in
+     do_query_repeat or due to a failure and an F: command) */
   if (seek_done)
     get_script_line( script, line );
   global_prior_command = drill_type;
@@ -968,7 +939,7 @@ do_drill( FILE *script, char *line ) {
 
 /*
   execute a timed speed test
- */
+*/
 static void 
 do_speedtest( FILE *script, char *line ) {
   int	errors = 0;		 /* error count */
@@ -1055,11 +1026,15 @@ do_speedtest( FILE *script, char *line ) {
 	      p--;		/* defeat p++ coming up */
 	      continue;
 	    }
-	  if ( c == ASCII_ESC ) break;
-      
+
+	  /* ESC is "give up"; ESC at beginning of exercise is "skip lesson"
+	     (this is handled outside the for loop) */
+	  if ( c == ASCII_ESC )
+	    break;
+
 	  /* check that the character was correct */
 	  if ( c == *p
-	       || ( cl_wpmode && c == ASCII_SPACE
+	       || ( cl_wp_emu && c == ASCII_SPACE
 		    && *p == ASCII_NL ))
 	    ADDCH( c );
 	  else 
@@ -1080,7 +1055,7 @@ do_speedtest( FILE *script, char *line ) {
 	    }
       
 	  /* perform any other word processor like adjustments */
-	  if ( cl_wpmode ) 
+	  if ( cl_wp_emu ) 
 	    {
 	      if ( c == ASCII_SPACE ) 
 		{
@@ -1115,63 +1090,73 @@ do_speedtest( FILE *script, char *line ) {
 	    }
 	}
   
-      /* look for escape key to "give up" */
-      if ( c == ASCII_ESC ) continue;
     
-      /* display timings */
-      end_time = (long)time( NULL );
-      display_speed( chars_typed, end_time - start_time,
-		     errors );
-      
-      /* check whether the error-percentage is too high (unless in s:) */
-      if (drill_type != C_SPEEDTEST_PRACTICE_ONLY &&
-	  (float)errors/(float)chars_typed > global_error_max/100.0) 
+      /* ESC not at the beginning of the lesson: "give up" */
+      if ( c == ASCII_ESC && chars_typed != 1)
+	continue; /* repeat */
+
+      /* skip timings and don't check error-pct if exit was through ESC */
+      if ( c != ASCII_ESC )
 	{
-	  sprintf( message, ERROR_TOO_HIGH_MSG, global_error_max );
-	  if ( ! wait_user( message, MODE_SPEEDTEST, script, line )) {
-	    seek_done = TRUE;
-	    break; /* function key */
-	  }
+	  /* display timings */
+	  end_time = (long)time( NULL );
+	  display_speed( chars_typed, end_time - start_time,
+			 errors );
+      
+	  /* check whether the error-percentage is too high (unless in s:) */
+	  if (drill_type != C_SPEEDTEST_PRACTICE_ONLY &&
+	      (float)errors/(float)chars_typed > global_error_max/100.0) 
+	    {
+	      sprintf( message, ERROR_TOO_HIGH_MSG, global_error_max );
+	      wait_user( message, MODE_SPEEDTEST );
 
-	  /* check for F-command */
-	  if (global_on_failure_label != NULL) {
-	    /* move to the label position in the file */
-	    if (fseek(script,global_on_failure_label->offset,SEEK_SET ) == -1)
-	      fatal_error( _("internal error: fseek"), NULL );
-	    global_line_counter = global_on_failure_label->line_count;
-	    /* tell the user about the misery :) */
-	    sprintf(message,SKIPBACK_VIA_F_MSG,global_on_failure_label->label);
-	    /* reset value unless persistent */
-	    if (!global_on_failure_label_persistent)
-	      global_on_failure_label = NULL;
-	    /* no need to check for fkey because seek_done is set anyway */
-	    wait_user( message, MODE_SPEEDTEST, script, line );
-	    seek_done = TRUE;
-	    break;
-	  }
+	      /* check for F-command */
+	      if (global_on_failure_label != NULL)
+		{
+		  /* move to the label position in the file */
+		  if (fseek(script,global_on_failure_label->offset,SEEK_SET )==-1)
+		    fatal_error( _("internal error: fseek"), NULL );
+		  global_line_counter = global_on_failure_label->line_count;
+		  /* tell the user about the misery :) */
+		  sprintf(message,SKIPBACK_VIA_F_MSG,
+			  global_on_failure_label->label);
+		  /* reset value unless persistent */
+		  if (!global_on_failure_label_persistent)
+		    global_on_failure_label = NULL;
+		  wait_user( message, MODE_SPEEDTEST );
+		  seek_done = TRUE;
+		  break;
+		}
 
-	  continue;
+	      continue;
+	    }
 	}
 
-      /* ask the user whether he/she wants to repeat */
-      if ( ! do_query_noscriptfile ( script, WANNA_REPEAT_MSG ) ) {
+      /* ask the user whether he/she wants to repeat or exit */
+      if ( c == ASCII_ESC && cl_no_skip ) /* honor --no-skip */
+	c = do_query_repeat (script, FALSE);
+      else
+	c = do_query_repeat (script, TRUE);
+      if (c == 'E') {
 	seek_done = TRUE;
-	break; /* function key */
+	break;
       }
-      if (global_resp_flag)
-	break; /* user has hit Y */
+      if (c == 'N')
+	break;
+
     }
 
   /* free the malloced memory */
   free( data );
 
+
   /* reset global_error_max */
   if (!global_error_max_persistent)
     global_error_max = cl_default_error_max;
 
-  /* buffer_command takes care of advancing `script' (and setting `line'),
-     so we only do it if an fkey was pressed (as a side-effect of a query/wait)
-  */
+  /* buffer_command takes care of advancing `script' (and setting
+     `line'), so we only do if seek_label had been called (in
+     do_query_repeat or due to a failure and an F: command) */
   if (seek_done)
     get_script_line( script, line );
   global_prior_command = C_SPEEDTEST;
@@ -1210,7 +1195,7 @@ do_clear( FILE *script, char *line ) {
 
 /*
   go to a label - the flag is used to implement conditional goto's
- */
+*/
 static void 
 do_goto( FILE *script, char *line, bool flag )
 {
@@ -1235,7 +1220,7 @@ do_goto( FILE *script, char *line, bool flag )
 
 /*
   exit from the program (implied on eof)
- */
+*/
 static void 
 do_exit( FILE *script ) 
 {
@@ -1249,28 +1234,110 @@ do_exit( FILE *script )
   exit( 0 );
 }
 
-/*
-  This is used to implement "builtin loops" in do_drill and do_speedtest
-  returns true if we just got the expected Y/N,
-  false if exit was by a function key
- */
-static bool
-do_query_noscriptfile( FILE *script, const char *prompt )
-{
-  char line[MAX_SCR_LINE];
 
-  sprintf(line, "Q:%s\n", prompt);
-  return do_query(script, line, FALSE);
+/*
+  Ask the user whether he/she wants to repeat, continue or exit
+  (this is used at the end of an exercise (drill/speedtest))
+  The second argument is FALSE if skipping a lesson isn't allowed (--no-skip).
+*/
+static char
+do_query_repeat ( FILE *script, bool allow_next )
+{
+  char resp;
+
+  /* display the prompt */
+  move( MESSAGE_LINE, 0 ); clrtoeol();
+  move( MESSAGE_LINE, COLS - strlen( MODE_QUERY ) - 2 );
+  ADDSTR_REV( MODE_QUERY );
+  move( MESSAGE_LINE, 0 );
+  if (allow_next)
+    ADDSTR_REV( REPEAT_NEXT_EXIT_MSG );
+  else
+    ADDSTR_REV( REPEAT_EXIT_MSG );
+
+  /* wait for [RrNnEe] (or translation of these) */
+  while (TRUE)
+    {
+      resp = getch_fl( ASCII_NULL );
+
+      if (toupper ((char)resp) == 'R' ||
+	  toupper ((char)resp) == RNE [0]) {
+	resp = 'R';
+	break;
+      }
+      if (allow_next && (toupper ((char)resp) == 'N' ||
+			 toupper ((char)resp) == RNE [2])) {
+	resp = 'N';
+	break;
+      }
+      if (toupper ((char)resp) == 'E' || toupper ((char)resp) == RNE [4]) {
+	if (fkey_bindings [11] != NULL &&
+	    do_query_simple (CONFIRM_EXIT_LESSON_MSG))
+	  {
+	    seek_label (script, fkey_bindings [11], NULL);
+	    resp = 'E';
+	    break;
+	  }
+	/* redisplay the prompt */
+	move( MESSAGE_LINE, 0 ); clrtoeol();
+	move( MESSAGE_LINE, COLS - strlen( MODE_QUERY ) - 2 );
+	ADDSTR_REV( MODE_QUERY );
+	move( MESSAGE_LINE, 0 );
+	if (allow_next)
+	  ADDSTR_REV( REPEAT_NEXT_EXIT_MSG );
+	else
+	  ADDSTR_REV( REPEAT_EXIT_MSG );
+      }
+    }
+
+  /* clear out the message line */
+  move( MESSAGE_LINE, 0 ); clrtoeol();
+  
+  return (char)resp;
 }
 
+
 /*
-  get a Y/N response from the user; get_next_script_line is FALSE if
-  this isn't started from a script-file Q:-command (as in
-  do_drill_noscriptfile) returns true if we just got the expected Y/N,
-  false if exit was by a function key
+  Same as do_query, but only used internally (doesn't set global_resp_flag,
+  returns the value instead) and doesn't accept Fkeys.
+  This is used to let the user confirm (E)xit.
+*/
+static bool
+do_query_simple ( char *text )
+{
+  char resp;
+
+  /* display the prompt */
+  move( MESSAGE_LINE, 0 ); clrtoeol();
+  move( MESSAGE_LINE, COLS - strlen( MODE_QUERY ) - 2 );
+  ADDSTR_REV( MODE_QUERY );
+  move( MESSAGE_LINE, 0 );
+  ADDSTR_REV( text );
+
+  /* wait for Y/N or translation of Y/N */
+  do
+    {
+      resp = getch_fl( ASCII_NULL );
+      
+      if (toupper (resp) == 'Y' || toupper (resp) == YN[0])
+	resp = 0;
+      if (toupper (resp) == 'N' || toupper (resp) == YN[2])
+	resp = -1;
+    }  while (resp != 0 && resp != -1);
+
+  /* clear out the message line */
+  move( MESSAGE_LINE, 0 ); clrtoeol();
+  
+  return resp == 0 ? TRUE : FALSE;
+}
+
+
+/*
+  get a Y/N response from the user: returns true if we just got the
+  expected Y/N, false if exit was by a function key
 */
 static bool 
-do_query( FILE *script, char *line, bool get_next_script_line )
+do_query( FILE *script, char *line )
 {
   int	resp;			/* response character */
   int	fkey;			/* function key iterator */
@@ -1283,7 +1350,7 @@ do_query( FILE *script, char *line, bool get_next_script_line )
   move( MESSAGE_LINE, 0 );
   ADDSTR_REV( SCR_DATA( line ) );
   
-  /* wait for a Y/N response, or matching FKEY */
+  /* wait for a Y/N response, translation of Y/N or matching FKEY */
   while ( TRUE ) 
     {
       resp = getch_fl( ASCII_NULL );
@@ -1339,9 +1406,8 @@ do_query( FILE *script, char *line, bool get_next_script_line )
   /* clear out the message line */
   move( MESSAGE_LINE, 0 ); clrtoeol();
 
-  /* get the next command (unless this is a "noscriptfile"-query) */
-  if (get_next_script_line)
-    get_script_line( script, line );
+  /* get the next command */
+  get_script_line( script, line );
 
   /* tell the caller whether we got Y/N or a function key */
   return ( ret_code );
@@ -1496,7 +1562,7 @@ do_on_failure_label_set( FILE *script, char *line )
 
 /*
   execute the directives in the script file
- */
+*/
 static void 
 parse_file( FILE *script, char *label ) {
 
@@ -1530,7 +1596,7 @@ parse_file( FILE *script, char *label ) {
 	case C_CLEAR:	do_clear( script, line ); break;
 	case C_GOTO:	do_goto( script, line, TRUE ); break;
 	case C_EXIT:	do_exit( script ); break;
-	case C_QUERY:	do_query( script, line, TRUE ); break;
+	case C_QUERY:	do_query( script, line ); break;
 	case C_YGOTO:	do_goto( script, line, global_resp_flag );
 	  break;
 	case C_NGOTO:	do_goto( script, line, !global_resp_flag );
@@ -1554,9 +1620,9 @@ parse_file( FILE *script, char *label ) {
 }
 
 /**
-  indent to column n
-  @param n is the amount of times to write ' ' (negative or zero means none)
- **/
+   indent to column n
+   @param n is the amount of times to write ' ' (negative or zero means none)
+**/
 static void
 indent_to( int n )
 {
@@ -1566,16 +1632,16 @@ indent_to( int n )
 }
 
 /**
-  Prints one usage option. 
-  The arguments op, lop and help should not have special characters '\n' '\t'
-  @param op is the short option
-  @param lop is the long option
-  @param help is the explanation of the option
-  @param col_op is the column where the short option will be displayed
-  @param col_lop is the column where the long option will be displayed
-  @param col_op is the column where the explanation will be displayed
-  @param last_col is the last column that can be used
- **/
+   Prints one usage option. 
+   The arguments op, lop and help should not have special characters '\n' '\t'
+   @param op is the short option
+   @param lop is the long option
+   @param help is the explanation of the option
+   @param col_op is the column where the short option will be displayed
+   @param col_lop is the column where the long option will be displayed
+   @param col_op is the column where the explanation will be displayed
+   @param last_col is the last column that can be used
+**/
 static void
 print_usage_item( char *op, char *lop, char *help, 
 		  int col_op, int col_lop, int col_help, int last_col ) 
@@ -1623,34 +1689,36 @@ print_usage_item( char *op, char *lop, char *help,
 
 /**
    Prints usage instructions
- **/
+**/
 static void
 print_help()
 {
   char *op[]= 
-  { "-e %",
-    "-n",
-    "-t",
-    "-f P",
-    "-c F,B",
-    "-s",
-    "-q",
-    "-l L",
-    "-w",
-    "-h",
-    "-v" };
+    { "-e %",
+      "-n",
+      "-t",
+      "-f P",
+      "-c F,B",
+      "-s",
+      "-q",
+      "-l L",
+      "-w",
+      "-k"
+      "-h",
+      "-v" };
   char *lop[]= 
-  { "--max-error=%",
-    "--notimer",
-    "--term-cursor",
-    "--curs-flash=P",
-    "--colours=F,B",
-    "--silent",
-    "--quiet",
-    "--start-label=L",
-    "--wpmode",
-    "--help",
-    "--version" };
+    { "--max-error=%",
+      "--notimer",
+      "--term-cursor",
+      "--curs-flash=P",
+      "--colours=F,B",
+      "--silent",
+      "--quiet",
+      "--start-label=L",
+      "--word-processor",
+      "--no-skip",
+      "--help",
+      "--version" };
   char *help[] = 
     { 
       _("default maximum error percentage (default 3.0); valid values are between 0.0 and 100.0"),
@@ -1662,6 +1730,7 @@ print_help()
       _("same as -s, --silent"),
       _("start the lesson at label 'L'"),
       _("try to mimic word processors"),
+      _("forbid the user to skip exercises"),
       _("print this message"),
       _("output version information and exit") };
   
@@ -1696,7 +1765,7 @@ print_help()
 
 /*
   parse command line options for initial values
- */
+*/
 static void 
 parse_cmdline( int argc, char **argv ) {
   int	c;				/* option character */
@@ -1711,7 +1780,8 @@ parse_cmdline( int argc, char **argv ) {
     { "silent",		no_argument, 0, 's' },
     { "quiet",		no_argument, 0, 'q' },
     { "start-label",	required_argument, 0, 'l' },
-    { "wpmode",		no_argument, 0, 'w' },
+    { "word-processor",	no_argument, 0, 'w' },
+    { "no-skip",        no_argument, 0, 'k' },
     { "help",		no_argument, 0, 'h' },
     { "version",	no_argument, 0, 'v' },
     { 0, 0, 0, 0 }};
@@ -1774,7 +1844,10 @@ parse_cmdline( int argc, char **argv ) {
 	  strcpy( cl_start_label, optarg );
 	  break;
 	case 'w':
-	  cl_wpmode = TRUE;
+	  cl_wp_emu = TRUE;
+	  break;
+	case 'k':
+	  cl_no_skip = TRUE;
 	  break;
 	case 'h':
 	  print_help();
@@ -1802,9 +1875,9 @@ parse_cmdline( int argc, char **argv ) {
 }
 
 
- /*
-   signal handler to stop curses stuff on intr
- */
+/*
+  signal handler to stop curses stuff on intr
+*/
 static void 
 catcher( int signal ) {
   
@@ -1818,7 +1891,7 @@ catcher( int signal ) {
 
 /*
   main routine
- */
+*/
 int main( int argc, char **argv ) 
 {
   WINDOW	*scr;				/* curses window */
@@ -1855,11 +1928,23 @@ This program is released under the GNU General Public License.");
     _("Your error-rate is too high. You have to achieve %.1f%%.");
   /* this message is displayed when the user has failed in a [DS]: drill,
      and an F:<LABEL> ("on failure label") is in effect */
-  SKIPBACK_VIA_F_MSG =
+  SKIPBACK_VIA_F_MSG=
     _("You failed this test, so you need to go back to %s.");
-  /* this is used for queries. you can translate the keys as well
-     (if you translate msgid "Y/N" accordingly) */
-  WANNA_REPEAT_MSG= _("Press Y to continue, N to repeat or Fkey12 to exit");
+  /* this is used for repeat-queries. you can translate the keys as well
+     (if you translate msgid "R/N/E" accordingly) */
+  REPEAT_NEXT_EXIT_MSG= 
+	_("Press R to repeat, N for next exercise or E to exit");
+  /* this is used for repeat-queries with --no-skip. you can translate
+     the keys as well (if you translate msgid "R/N/E" accordingly) */
+  REPEAT_EXIT_MSG= 
+	_("Press R to repeat or E to exit");
+  /* This is used make the user confirm (E)xit in REPEAT_NEXT_EXIT_MSG */
+  CONFIRM_EXIT_LESSON_MSG=
+    _("Are you sure you want to exit this lesson? [Y/N]");
+  /* This message is displayed if the user tries to skip a lesson
+     (ESC ESC) and --no-skip is specified */
+  NO_SKIP_MSG= 
+    _("Sorry, gtypist is configured to forbid skipping exercises.");
   /* this must be adjusted to the right with one space at the end.
      Leading whitespace is important because it is displayed in reverse
      video because it must be aligned with the next two messages 
@@ -1875,8 +1960,10 @@ This program is released under the GNU General Public License.");
      video and because it must be aligned with the next last two messages
      (it's best to run gtypist to see this in practice)   */
   SPEED_PCT_ERROR=_("            with %.1f%% errors ");
-  /* this is used to translate the keys for Y/N-queries. Must be two uppercase
-     letters. Y/N will still be accepted as well */
+  /* this is used to translate the keys for Y/N-queries. Must be two
+     uppercase letters separated by '/'. Y/N will still be accepted as
+     well. Note that the messages (prompts) themselves cannot be
+     translated because they are read from the script-file. */
   YN = _("Y/N");
   if (strlen(YN) != 3 || YN[1] != '/' || !isupper(YN[0]) || !isupper(YN[2]))
     {
@@ -1885,6 +1972,19 @@ This program is released under the GNU General Public License.");
 	       argv0, YN );
       exit( 1 );
     }
+  /* this is used to translate the keys for Repeat/Next/Exit
+     queries. Must be three uppercase letters separated by slashes. */
+  RNE = _("R/N/E");
+  if (strlen(RNE) != 5 ||
+      !isupper(RNE[0]) || RNE[1] != '/' ||
+      !isupper(RNE[2]) || RNE[3] != '/' ||
+      !isupper(RNE[4]))
+    {
+      fprintf( stderr,
+	       "%s: i18n problem: invalid value for msgid \"R/N/E\": %s\n",
+	       argv0, RNE );
+      exit( 1 );
+    }
   
   /* get our name for error messages */
   argv0 = argv[0] + strlen( argv[0] );
@@ -1971,4 +2071,3 @@ This program is released under the GNU General Public License.");
   /* for lint... */
   return( 0 );
 }
-
